// Ported from https://github.com/slush0/stratum-mining/blob/master/lib/merkletree.py
var util = require('./util.js');

var MerkleTree = module.exports = function MerkleTree(data) {
    function merkleJoin(h1, h2) {
        return util.sha256d(Buffer.concat([h1, h2]));
    }

    function calculateSteps(data) {
        var steps = [];
        var sequence = data;
        var sequenceLen = sequence.length;
        while (sequenceLen > 1) {
            steps.push(sequence[1]);
            if (sequenceLen % 2) sequence.push(sequence[sequenceLen - 1]);
            var newSequence = [null];
            for (var i = 2; i < sequenceLen; i += 2) {
                newSequence.push(merkleJoin(sequence[i], sequence[i + 1]));
            }
            sequence = newSequence;
            sequenceLen = sequence.length;
        }
        return steps;
    }

    // Used to calculate merkle root without adding a coinbase later
    function calculateRoot(_data) {
        var data = _data; // We dont want to work in-place
        // This is a recursive function
        if (data.length > 1) {
            if (data.length % 2 !== 0)
                data.push(data[data.length - 1]);
            // Hash
            var newData = [];
            for (var i = 0; i < data.length; i += 2) newData.push(merkleJoin(data[i], data[i + 1]));
            return calculateRoot(newData);
        }
        return data[0];
    }
    this.data = data;
    this.steps = calculateSteps(data);
    this.root = calculateRoot(data[0] == null ? data.slice(1) : data);
}

MerkleTree.prototype = {
    withFirst: function(f) {
        this.steps.forEach(function(s) {
            f = util.sha256d(Buffer.concat([f, s]));
        });
        return f;
    },
    // Used to develop steps to prove a single hash is part of a merkle root
    getHashProof: function(h) {
        var data = this.data;
        if (data.length == 1) return Buffer.concat([util.varIntBuffer(0), util.packInt32LE(0)]);
        var ind = data.indexOf(h);
        if (ind < 0) return undefined; // Cant prove; it is not part of this merkle tree
        var branchLen = 0;
        var hashBuffer = new Buffer(0);
        var sideMask;
        for (; data.length > 1; branchLen++) {
            if (data.length % 2 !== 0)
                data.push(data[data.length - 1]);
            if (ind % 2 === 0) {
                // We need right side
                Buffer.concat([hashBuffer, data[ind + 1]]);
                // No need to write side mask because it should already be 0
            } else {
                // We need left side
                Buffer.concat([hashBuffer, data[ind - 1]]);
                sideMask = sideMask & (1 << branchLen);
            }
            // Calculate the next level of the merkle root.
            var newData = [];
            for (var i = 0; i < data.length; i += 2) {
                newData.push(merkleJoin(data[i], data[i + 1]));
            }
            data = newData;
            ind = Math.floor(ind / 2);
        }
        branchLen++;
        return Buffer.concat([util.varIntBuffer(branchLen), hashBuffer, util.serializeNumber(sideMask)]);
    }
};
